#include <dbsim.h>
#include <object.h>
#include <disk.h>
#include <op.h>
#include <query.h>
#include <language.h>
#include <strings.h>
#include <math.h>

// ---------------------------------------------------------------------------
// class implementation for Disk (disk i/o facility)
// ---------------------------------------------------------------------------

Disk::Disk(): facility("disk")
{
    strcpy(dname, "disk");
    curCylinder = 0;
    curDirection = IN;
    servStat = new table("disk-serv");
    waitStat = new qtable("disk-wait");
    nextCleanPage = 0;
    freeExtents = NULL;
    pendingHandoff = FALSE;
}

Disk::Disk(char *s): facility(s)
{
    strcpy(dname, s);
    curCylinder = 0;
    curDirection = IN;
    servStat = new table(s);
    waitStat = new qtable(s);
    nextCleanPage = 0;
    freeExtents = NULL;
    pendingHandoff = FALSE;
}

Disk::~Disk()
{
}

void Disk::use(int pageId)	// satisfy C++ with DISK_READ default
{
    use(pageId, DISK_READ, NO_PREFETCH);
}


extern "C" void exit(int);

void Disk::use(int pageId, int reqType, int prefetch, Op* op)
{
    DiskReq* newReq;		// descriptor for this disk request
    DiskReq* nextReq;		// descriptor for next disk request
    int seekDist;		// number of cylinders to seek
    float diskTime;		// total time required for actual use of disk
    int nextCylPage1;		// first page number of next cylinder
    int cylPages;		// number of pages remaining in this cylinder
    int xtraPages;		// number of pages to prefetch in a read
    CacheContext* context;	// cache context for the current request
    float respTime;     	// resp time for this request (incl. queueing)

    // note entry and allocate/initialize disk request descriptor

    waitStat->note_entry();
    respTime = simtime();   // note req start time
    newReq = new DiskReq(pageId);

    // Records disk reads in operator, if present

    if (op != NULL && reqType == DISK_READ) op->cnt_diskRead++; 

    if (Flags[TRACE_DISK]) {
        printf("%.3f(%d) i/o REQUEST [%s]: page = %d\n", 
	       clock, identity(), dname, pageId);
    }

    // if disk is currently busy, then queue the request and wait

    if (facility::status() == BUSY || pendingHandoff) {
	// disk is busy, so queue the incoming request
        if (curCylinder > newReq->cylinder) {
	    // inner cylinder requests handled on the way in
	    inQueue.enqueueIn(newReq);
        } else if (curCylinder < newReq->cylinder) {
	    // outer cylinder requests handled on the way out
	    outQueue.enqueueOut(newReq);
        } else if (curCylinder == newReq->cylinder) {
	    // requests for current cylinder handled first
	    if (curDirection == IN) {
	        inQueue.enqueueIn(newReq);
	    } else { // curDirection == OUT
	        outQueue.enqueueOut(newReq);
	    }
        }
        newReq->ioReady->wait();
	pendingHandoff = FALSE;
    }

    // now use the disk facility for the appropriate amount of time,
    // figured using the Bitton/Gray formula, and then dispose of the
    // request and note that it has been served;  also account for a
    // multi-page cache with a write-through protocol for updates (also
    // updating the cached copy of the page, if any)

    facility::reserve();
    
    if (Flags[TRACE_DISK]) {
        printf("%.3f(%d) i/o START [%s]: dir = %s, cyl = %d, new cyl = %d\n", 
                clock, identity(), dname, headDir[curDirection], curCylinder,
		newReq->cylinder);
    }

    if ((reqType == DISK_WRITE) || (DiskCachePages*DiskCacheContexts == 0)) {

	// caching doesn't save time for writes or if cache size is zero

        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o cache unused/invalidated [%s]\n", 
                   clock, identity(), dname);
        }
        seekDist = abs(newReq->cylinder - curCylinder);
        curCylinder = newReq->cylinder;	// note new head position
        diskTime = (int) ( (SeekFactor * sqrt((double) seekDist)) +
                    uniform(MinLatency, DiskRotation) + SettleTime + XferTime );
        hold(diskTime);
        servStat->record(diskTime);

	// NOTE: cache contents are left undisturbed in this case!

	// Records disk arm resp time on a job class basis (too expensive to 
        // record at any finer granularity, and still have accurate rollup
        // to higher levels)

	if (op != NULL) op->query->jobClass->armResp->record(diskTime);

    } else if (cache.whichContext(newReq->page) == NULL) {

	// a cache miss has occurred for a read request

        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o cache miss [%s], ", clock, identity(), dname);
	    cache.print();
	    printf("\n");
	    
        }

	// first compute how many pages are to be prefetched (as xtraPages)

	if (prefetch == PREFETCH) {
	    // may prefetch up through end of current cylinder
	    nextCylPage1 = (newReq->cylinder + 1) * PagesPerCyl;
	    cylPages = nextCylPage1 - (newReq->page + 1);
	    xtraPages = (DiskCachePages < cylPages) ? DiskCachePages : cylPages;
	} else {
	    // no prefetch requested for this read operation
	    xtraPages = 0;
	}

	// now go ahead and read one plus xtraPages disk pages

        seekDist = abs(newReq->cylinder - curCylinder);
        curCylinder = newReq->cylinder;	// note new head position
        diskTime = (int) ( (SeekFactor * sqrt((double) seekDist)) +
                    uniform(MinLatency, DiskRotation) + SettleTime +
		    (1 + xtraPages) * XferTime );
        hold(diskTime);
        servStat->record(diskTime);

	// cache must be updated to contain any/all prefetched pages

	if (xtraPages > 0) {
	    context = cache.nextCacheContext();
	    context->firstCachedPage = newReq->page + 1;
	    context->lastCachedPage  = newReq->page + xtraPages;
	}

        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o [%s] new context, ", clock, identity(), dname);
	    cache.print();
	    printf("\n");
        }

	// Records seek distances on a job class basis (too expensive to 
        // record at any finer granularity, and still have accurate rollup
        // to higher levels)

	if (op != NULL) op->query->jobClass->armResp->record(diskTime);

    } else {

        // a cache hit has occurred for a read request

        if (Flags[TRACE_DISK]) {
            printf("%.3f(%d) i/o cache hit [%s], ", clock, identity(), dname);
	    cache.print();
	    printf("\n");
        }

	// Records disk cache hits in operator, if present
	
	if (op != NULL) op->cnt_diskCacheHit++; 

    }

    if (Flags[TRACE_DISK]) {
        printf("%.3f(%d) i/o DONE [%s], dir = %s, cyl = %d\n",
                clock, identity(), dname, headDir[curDirection], curCylinder);
    }

    facility::release();
    delete newReq;
    waitStat->note_exit();

    // Records total request resp time on a job class basis

    respTime = simtime() - respTime; // calc request resp time
    if (respTime < 0.0) respTime = 0.0; // patch for CSIM BUG in facil::reserve
    if (op != NULL) op->query->jobClass->diskResp->record(respTime);

    // finally, pass the disk to the next requester, if any

    if (curDirection == IN) {
        // disk head is currently moving in
	if (inQueue.nonempty()) {
	    // get next request in current direction
	    nextReq = inQueue.dequeue();
	} else if (outQueue.nonempty()) {
	    // change direction and get next request
	    curDirection = OUT;
	    nextReq = outQueue.dequeue();
	} else {
	    nextReq = NULL;
	}
    } else {  // curDirection == OUT
	// disk head is currently moving out
	if (outQueue.nonempty()) {
	    // get next request in current direction
	    nextReq = outQueue.dequeue();
	} else if (inQueue.nonempty()) {
	    // change direction and get next request
	    curDirection = IN;
	    nextReq = inQueue.dequeue();
	} else {
	    nextReq = NULL;
	}
    }
    if (nextReq != NULL) {
	// notify next requester of disk availability
	pendingHandoff = TRUE;
        nextReq->ioReady->set();
    }
}

int Disk::AllocExtent()
{
    int firstPage;			// starting page of new extent
    Extent* ext;			// descriptor for a free extent
    if (freeExtents == NULL) {
	// allocate a brand new extent
	firstPage = nextCleanPage;
	nextCleanPage += EXTENT_SIZE;

    } else {
	// take a previously used extent from the free list
	ext = freeExtents;
	freeExtents = freeExtents->nextFree;
	firstPage = ext->firstPage;
	delete ext;
    }
    return(firstPage);
}

void Disk::FreeExtent(int firstPage)
{
    Extent* ext = new Extent;		// get a new extent descriptor,
    ext->firstPage = firstPage;		// note the id of its first page,
    ext->nextFree = freeExtents;	// and add it to the free list
    freeExtents = ext;
}

void Disk::print() 
{
    printf("\tDisk [%s]:\n\t\t curr cyl %d, curr direction %s, next clean pg %d\n\t\twaitStat qtable @  X'%x', freeExtent list @ X'%x'\n", dname, curCylinder, headDir[curDirection], nextCleanPage, waitStat, freeExtents);
    printf("\t\tIn queue:");
    inQueue.print();
    printf("\n\t\tOut queue:");
    outQueue.print();
    printf("\n\t\t");
    cache.print();
    printf("\n");
    fflush(stdout);
}    

extern "C" double sqrt(double);

void Disk::printStats()
{
    printf("%5.3f %7.3f %7.3f %7.3f %7.3f %7.3f %5.3f %7d %7d %5.2f %4d\n",
	   util(),  // utilization "at controller"
	   resp(),  // or waitStat->qtable_qtime() gives same answer
	   servStat->table_mean(), // mean svc time "at disk"
	   servStat->table_min(),
	   servStat->table_max(),
	   (float) sqrt((double) servStat->table_var()), // std dev disk svc
	   
	   // completions() counts requests received by the "controller",
	   // servStat->table_cnt() counts requests that miss the cache
	   // and are serviced by the disk, the difference = # cache hits,
	   // ratio of cache hits to controller reqs = cache hit rate
	   
	   (completions() == 0) ? 0.0 :
	   (float) (completions() - servStat->table_cnt()) / completions(),
	   
	   completions(),             // "controller" completions
	   servStat->table_cnt(),     // "disk" completions
	   waitStat->qtable_qlen(),   // mean queue len
	   waitStat->qtable_max()     // max queue len
	   );
}	 



