///////////////////////////////////////////////////////////////////////////////
//
//  Name:        nodemgr.c
//
//  Title:       Node manager process.
//
//  Function:    Manages node resources:  right now only operator processes.
//                 
//  Change Activity:
//
///////////////////////////////////////////////////////////////////////////////

#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "msg.h"
#include "mailbox.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"

void processSelect(Node*, QueryMsg*);
void processScan(Node*, SelectPlan*, Mailbox*);
void processStore(Node*, QueryMsg*);
void processTerm(Node*, QueryMsg*);
void processJoin(Node*, QueryMsg*);
void selectOperator(SelectOp*);
void scanOperator(ScanOp*);
void storeOperator(StoreOp*); 
void terminalOperator(TermOp*); 
void joinOperator(JoinOp*); 
File*  createFile(int, int, int, int);
SplitTable* copySplitTable(SplitTable*);

void nodeMgr(Node* n) 
{
    Msg*	msg;
    QueryMsg*	qmsg;
    Mailbox*	inPort;
  
    create("NodeMgr");    

    if (Flags[TRACE_CONTROL])
      printf("NodeMgr is running\n");
    inPort = n->nodeMgrPort;

    while (SimRunning)
    {
        msg = inPort->recv();     // wait for a message
        switch (msg->type) 
	{               
      	    case QUERYPLAN: 
		qmsg = (QueryMsg*) msg;

		if (Flags[TRACE_QUERY]) {
		  printf("NodeMgr %d, received a query:\n", n->id);
		  qmsg->query->print(0);
		  printf("\n\n");
		}

		// now start a proper process to execute the operation
		switch (qmsg->query->opCode)
		{
		    case (SELECT):  
			processSelect(n, qmsg);
			break;
		    case (STORE):  
			processStore(n, qmsg);
			break;
		    case (TERMOP):  
			processTerm(n, qmsg);
			break;
		    case (JOIN):  
			processJoin(n, qmsg);
			break;
		    default: 
			printf("unimplemented operation %d\n",
				qmsg->query->opCode);
		}
		(void) QueryMsg::free(qmsg);
                break;

	    default:
		printf("NodeMgr %d received an illegal message type\n", n->id);
        }
    }
}

void processSelect(Node* n, QueryMsg* msg)
{
    SelectOp*	selectOp;
    SelectPlan* qry;

    qry = (SelectPlan *) msg->query; 

    selectOp = new SelectOp(n);
    selectOp->query = qry ;
    selectOp->transId = qry->transId;
    selectOp->initiator = msg->sourceBox;
    selectOp->outTupSize = qry->outTupSize;
    selectOp->inTupSize = qry->inTupSize;
    selectOp->indexSelectivity = qry->indexSelectivity;

    // if xmed operator 
    //      get file name and new split table from the query message
    // else
    //      get file and split table from the query plan node
    if (qry->selectType != REMOTE)
    {
	selectOp->splitTable = copySplitTable(msg->splitTable);
	selectOp->splitTable->tupleSize = qry->outTupSize;
	selectOp->file = msg->file;
	selectOp->index = msg->index;
	delete msg->splitTable;
    }
    else // intermediate select
    {
	selectOp->splitTable = copySplitTable(qry->splitTable);
	selectOp->splitTable->tupleSize = qry->outTupSize;
	selectOp->file = qry->file;
	selectOp->index = qry->index;
    }

    selectOp->selFactor = qry->selFactor;
    
    // start the select operator
    selectOperator(selectOp); 
}

void processScan(Node* n, SelectPlan* sp, Mailbox* inPort)
{
    ScanOp*	scanOp;

    scanOp = new ScanOp(n);
    scanOp->query = sp ;
    scanOp->transId = sp->transId;
    scanOp->initiator = inPort;
    scanOp->outTupSize = sp->outTupSize;
    scanOp->inTupSize = sp->inTupSize;
    scanOp->indexSelectivity = sp->indexSelectivity;
    scanOp->file = sp->file;
    scanOp->index = sp->index;

    scanOp->splitTable = new SplitTable(1, sp->outTupSize);
    scanOp->splitTable->splitArray[0].numTuples = 0;
    scanOp->splitTable->splitArray[0].destNode = inPort;
    
    // start the select operator
    scanOperator(scanOp); 
}

void processStore(Node* n, QueryMsg* msg)
{
    StoreOp*	storeOp;
    StorePlan*  qry;

    qry = (StorePlan *) msg->query; 
    storeOp = new StoreOp(n);

    storeOp->query = qry;
    storeOp->transId = qry->transId;
    storeOp->initiator = msg->sourceBox;
    storeOp->outTupSize = qry->outTupSize;
    storeOp->inTupSize = qry->inTupSize;
    storeOp->inStreams = qry->inStreams;

    // if the store type is REMOTE (store for final result of query)
    //       get file id from the query node plan
    // else
    //       make a new xmed file
    //       file name is sent back to the scheduler when store completes
    //           to be used by the xmed select for this bucket
    if (qry->storeType == REMOTE)
    {
      storeOp->file = qry->file;
      storeOp->diskId = qry->diskId;
    }
    else // xmed operator
    {
      storeOp->diskId = random(0, DisksPerNode - 1);
      storeOp->file = new File(qry->transId, 1, &n, 1, &(storeOp->diskId), 
                               0, storeOp->outTupSize);
      if (Flags[TRACE_OP])
        printf("New xmed file created for store on node %d",n->id);

    }

    // start the storeOperator
    storeOperator(storeOp); 
}

void processTerm(Node* n, QueryMsg* msg)
{
    TermOp*	termOp;
    TermPlan*  qry;

    qry = (TermPlan *) msg->query; 
    termOp = new TermOp(n);

    termOp->query = qry;
    termOp->transId = qry->transId;
    termOp->initiator = msg->sourceBox;
    termOp->inStreams = qry->inStreams;
    // start the terminal operator
    terminalOperator(termOp); 
}

void processJoin(Node* n, QueryMsg* msg)
{
    JoinOp*	joinOp;
    JoinPlan* 	qry;

    qry = (JoinPlan *) msg->query; 
    joinOp = new JoinOp(n);

    joinOp->query = qry;
    joinOp->transId = qry->transId;
    joinOp->initiator = msg->sourceBox;
    joinOp->outTupSize = qry->outTupSize;
    joinOp->inTupSizeR = qry->inTupSizeR;
    joinOp->inTupSizeL = qry->inTupSizeL;
    joinOp->inStreamsR = qry->inStreamsR;
    joinOp->inStreamsL = qry->inStreamsL;
    joinOp->selFactor = qry->selFactor;
    joinOp->pagesAllocated = qry->maxPagesNeeded;

    // 15 for padding and alignment ???
    joinOp->tupsPerNetPageInR = NETPKTSIZE/(qry->inTupSizeR + 15); 
    joinOp->tupsPerNetPageInL = NETPKTSIZE/(qry->inTupSizeL + 15); 
    joinOp->tupsPerNetPageOut = NETPKTSIZE/(qry->outTupSize + 15); 

    // start the join operator
    joinOperator(joinOp); 
}




