#include "dbsim.h"
#include "object.h"
#include "mailbox.h"
#include "msg.h"
#include "node.h"
#include "bufmgr.h"
#include "split.h"
#include "language.h"
#include "db.h"
#include "op.h"
#include "query.h"
#include "cost.h"

extern "C" {
    void exit(int);
}


// routine to simulate a scan operation operating on a 
// disk-resident relation

void scanOperator(ScanOp* op)
{
    int 	pageId, diskId, numTuples, fullTuples;
    float       tuplesProduced; // running sum of tuples produced
    Node* 	n;
    int 	e;	
    CntlMsg*	cMsg;
    int		instCnt;
    int		tupleCnt;  // number of tuples processed by operator
    TIME	startTime; // time operator starts running
    int         scanType;       // index or file scan
    const int   FILE_SCAN = 1;
    const int   INDEX_SCAN = 2;
    int		prefetchData;	// whether or not to prefetch data pages
    int         scanRC;         // file/index scan return code
    int         pageType;       // index scan page type returned from nextPage
    int         fileId;         // file to be scanned (direct or via index)

    // following four variables are for performance reasons only
    int		tupsPerNetPageOut;
    IndexScan*	indexScan;	// scan for input file
    FileScan*	fileScan;	// scan for input file
    Mailbox*	inPort;		// mailbox for process to receive messages
    SplitTable*	splitTable;	// local pointer for performance
    int         splitIdx;     
    int		selectType;	// local value of selectType 
    int         totalPages;     // number of pages filled in the buffer

    create("scan");

    n = op->node;
    // construct mail box for process to recieve messages
    inPort = new Mailbox("opMbox", n, op);

    // initializations 
    tupleCnt = 0;     // total number of tuples processed
    startTime = clock;

    tupsPerNetPageOut = NETPKTSIZE/(op->outTupSize + 15);


    // extract the selectType.  REMOTE means the select is a normal
    // select. XMED means a select reading an intermediate file
    selectType = ((SelectPlan*) op->query)->selectType;

    // set up a scan of the input file
    if (op->index == NULL) {     // no index means a file scan
      scanType = FILE_SCAN;
      prefetchData = PREFETCH;
      // initiate the actual file scan
      fileScan = new FileScan(op->file, n);
      pageType = DATA_PAGE;      // always the case for a file scan
      fileId = op->file->fileNo;
      scanRC = fileScan->nextPage(pageId, diskId, numTuples);
    }
    else {
      scanType = INDEX_SCAN;
      if (op->index->indexType == CLUSTERED) {
          prefetchData = PREFETCH;
      } else {
	  prefetchData = NO_PREFETCH;
      }
      // initiate the actual index scan
      indexScan = new IndexScan(op->index, n, op->indexSelectivity);
      fileId = op->index->data->fileNo;
      scanRC = indexScan->nextPage(pageId, pageType, diskId, numTuples);
    }
    
    splitTable = op->splitTable;


    
    if (Flags[TRACE_OP])
    {
      printf("scan (%d) started on node=%d, selecttype = %d, scantype=%d, file=%d, outtupsize=%d\n", identity(),
	     n->id,selectType, scanType, fileId, op->outTupSize);
    }
    
    // Note that no CPU instructions are consumed to start/end a scan op, since
    // in a real system, scan processes are started once per disk, and just
    // sit there waiting.  They are not started on demand in response to
    // any query.

    tuplesProduced = 0.0;
    totalPages = 0;

    while (scanRC != END_OF_SCAN) // continue until end of file
    {

        if (totalPages >= 2)
	{
	    if (totalPages > 2)
	    {
	        printf("ERROR: scan: totalpages (%d) > 2\n", totalPages);
		exit(NOTOK);
	    }
	    cMsg = (CntlMsg*) inPort->recv();
	    if (cMsg->opCode != PAGEDONE)
	    {
                 printf("scan: - node %d, illegal cntl msg code =%d\n",
	    	        n->id, cMsg->opCode);
		 exit(NOTOK);
            }
	    totalPages--;
	    (void) CntlMsg::free(cMsg);
	}
	      
//        if (Flags[TRACE_OP])
//	  printf("scan (%d), node %d, pageId %d, pageType %d, pagedisk %d, numTuples %d\n", identity(),n->id, pageId, pageType, diskId, numTuples);

	// love all the index pages and hate all the data pages
	//   assumes that we do not want to see the same data page twice
	if (pageType == DATA_PAGE)
	    e = n->bufMgr->ReadPage(diskId, pageId, op->file, HATE,
		                          prefetchData, op);
	else
	    e = n->bufMgr->ReadPage(diskId, pageId, op->file, LOVE,
		  			  NO_PREFETCH, op);
	if (e != OK) printf("in Select, bufMgr.readpage returned error\n");

	// consume some CPU time as appropriate for this page
        if (pageType == INDEX_PAGE) {
            instCnt = numTuples * (INST_INDEX_ENTRY_TEST);
            numTuples = 0; // skip tuple loop and avoid excessive indentation!
        }
        else
	    instCnt = numTuples * INST_TREAD;

        n->cpu.use(((float) instCnt) / (Mips * 1000)); //milliseconds
        op->cnt_CPUInstr += instCnt;

        tuplesProduced += numTuples;


        if (tuplesProduced >= 1.0)
        {

	    fullTuples = (int) tuplesProduced;

	    splitIdx = 0;

	    splitTable->splitArray[splitIdx].numTuples += fullTuples;

	    if ((splitTable->splitArray[splitIdx].numTuples)
		>= tupsPerNetPageOut)
	    {
	        totalPages++;

	        // split buffer is full, send off buffer to destination
//		if (Flags[TRACE_OP])
//	           printf("scan (%d) on node %d sending netpage (%d tuples)\n",
//	  	    identity(), n->id, 
//	 	    splitTable->splitArray[splitIdx].numTuples);

		splitTable->sendNetBuffer(splitIdx, FALSE, n, inPort, CHEAP);
            }

	    //update the count of tuples produced so far and the tuples 
	    // produced
	    tupleCnt += fullTuples; 
	    tuplesProduced -= fullTuples;
	}

	// finished with page, unfix it!
	e = n->bufMgr->UnfixPage(diskId, pageId);
	if (e != OK) 
	        printf("in Select, bufMgr.unfixpage returned error\n");

        // get next page to process
        if (scanType == FILE_SCAN) 
            scanRC = fileScan->nextPage(pageId, diskId, numTuples);
	else
            scanRC = indexScan->nextPage(pageId, pageType, diskId, numTuples);

    } // for all pages
  
    // exhausted input file  flush all output buffers
    splitTable->flushNetBufs(n, inPort, CHEAP);
    totalPages++;

    if (Flags[TRACE_OP])
      printf("scan (%d) on node %d finishing time %f\n", 
             identity(),n->id, clock);

    // finally, if the file is a "LOCAL" select delete the input file 
    if (selectType == XMED) {
	// first flush the pages from the buffer pool.  Since
	// they are clean this will incur no I/O but it will cause
	// all the pages to be marked invalid
        if (Flags[TRACE_OP])
	    printf("scan (%d) calling local buffer flush followed delete\n", 
            identity());

        e = n->bufMgr->FlushFile(op->file);
	if (e != OK) printf("Scan, bufMgr->FlushFile returned error %d\n",e);
	op->file->deleteFrag(n, diskId);
    }



    while (totalPages > 0)
    {
	cMsg = (CntlMsg*) inPort->recv();
	if (cMsg->opCode != PAGEDONE)
	{
	    printf("scan: - node %d, illegal cntl msg code =%d\n",
		   n->id, cMsg->opCode);
	    exit(NOTOK);
	}
	totalPages--;
	(void) CntlMsg::free(cMsg);
    }

    // send opdone message to the parent select operator
    cMsg = CntlMsg::alloc();
    cMsg->opCode = OPDONE;
    cMsg->send (op->initiator, inPort);
    

    if (scanType == FILE_SCAN) delete fileScan; // delete the file scan
    else delete indexScan; // delete index scan
    delete inPort;
    delete op;	// delete the select operator object

}
