
///////////////////////////////////////////////////////////////////////////////
//
//  Name:        msg.c
//
//  Function:    Contains the message class and subclass (DataMsg, QueryMsg,
//               CntlMsg) method implementations.
//                 
//  Change Activity:
//
//     08/25/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

#include "dbsim.h"
#include "object.h"
#include "node.h"
#include "op.h"
#include "msg.h"
#include "mailbox.h"
#include "terminal.h"
#include <math.h>


extern "C" {
  double floor(double x);
  void exit(int);
}

int Msg::nextSeqNum = 0;

QueryMsg* QueryMsg::freeMsg = NULL; // no free query msgs initially
DataMsg* DataMsg::freeMsg = NULL; // no free data msgs initially
CntlMsg* CntlMsg::freeMsg = NULL; // no free control msgs initially

///////////////////////////////////////////////////////////////////////////////
//
//  Msg::Msg       Initializes message fields to detectably invalid values.
//
///////////////////////////////////////////////////////////////////////////////

Msg::Msg() {
  id = ++nextSeqNum;                 // assigns a seq number to the Msg
  sourceBox = destBox = (Mailbox*) NULL;
  type =  numBytes = -1;
  timestamp = 0.0;
}

///////////////////////////////////////////////////////////////////////////////
//
//  Msg::send      Superclass send.  If the sender is a terminal, just does
//                 a CSIM send.  If the sender is a node, charges the
//                 sending node's CPU for the message protocol pathlength.
//                 If the message short circuits the network, also charges 
//                 the sending node's for the mem to mem copy cost.  The 
//                 receiver is charged for any wire delay in Mailbox::receive.
//
///////////////////////////////////////////////////////////////////////////////

void Msg::send(Mailbox* dest, Mailbox* source) {
    Node* sourceNode;
    int numInstr;

    // Copies addessing info to state data

    sourceBox = source;
    destBox = dest;
 
    if (sourceBox->type == TERMINAL) 
    {
	// message is from a terminal; just send it

    	destBox->send((int) this);
	if (Flags[TRACE_MSG]) 
	  printf("Msg::send id %d, src %x, dest %x, %d bytes - from terminal\n",
		 id, sourceBox, destBox, numBytes);

    	return;
    }
    else 
    {
	// message is from a node

	sourceNode = sourceBox->owningNode;
 
  	// Accounts for protocol costs on the sender's CPU
	// times are in milliseconds
  	sourceNode->cpu.use(((float) PROTOCOLCOST) / (Mips * 10e3)); 
        if (sourceBox->owningOp != NULL) 
          sourceBox->owningOp->cnt_CPUInstr += PROTOCOLCOST;

  	// If packet shorts curcuits the network, 
	//  simulates a mem to mem copy cost
  	if (sourceNode == destBox->owningNode) 
	{
    	    if (numBytes <= 0) {
      	    	printf("Msg::send - bad msg size (%d)\n", numBytes);
      		numInstr = 800;
    	    }
            else
                numInstr = (int) floor( 1.181 * numBytes + 773.0);

////////// original set of numbers which the above formula approximates
//    	    else if (numBytes <= 56) numInstr = 879;
//    	    else if (numBytes <= 136) numInstr = 969;
//    	    else if (numBytes <= 286) numInstr = 1182;
//    	    else if (numBytes <= 536) numInstr = 1449;
//    	    else if (numBytes <= 786) numInstr = 1749;
//    	    else if (numBytes <= 1036) numInstr = 2016;
//    	    else if (numBytes <= 1536) numInstr = 2808;
//    	    else if (numBytes <= 2036) numInstr = 3174;
//    	    else if (numBytes <= 3036) numInstr = 4338;
//    	    else if (numBytes <= 5036) numInstr = 6747;
//    	    else if (numBytes <= 8192) numInstr = 10401;
//    	    else if (numBytes > 8192)
//            {
//              printf("msg::send - too many bytes (%d), msg id %d\n", 
//			numBytes, id);
//      	      numInstr = 14000;
//    	    }

    	   if (Flags[TRACE_MSG]) 
      	     printf("Msg::send - msg(%d) short circuited, %d bytes, %d instr\n",
             id, numBytes, numInstr);

     	   sourceNode->cpu.use(((float) numInstr) / (Mips * 10e3));
           if (sourceBox->owningOp != NULL) 
             sourceBox->owningOp->cnt_CPUInstr += numInstr;
  	}
        else
        {
   	    // Otherwise, the packet will actually need to use the network.
            // In this case, the receiver must experience the wire delay, since
            // otherwise the message send operation will be synchronous and
            // may prevent pipelining between a sending and receiving process.
            // The wire delay code is in Mailbox::receive.

            // timestamp the message so the reciever can find out
            // if it needs to wait before processing the message

            timestamp = clock;
        }

        
        // In either case (short circuit or not), we must pass the message 
        // to the appropriate CSIM process to pick up.
 
        destBox->send((int) this);    // actually send msg to csim process
    }
}


///////////////////////////////////////////////////////////////////////////////
//
//  Msg::cheapSend  models cheap interprocess messages which do not use
//                  the message subsystem. can be used only for nodes
//    
///////////////////////////////////////////////////////////////////////////////

void Msg::cheapSend(Mailbox* dest, Mailbox* source) {
    Node* sourceNode;

    // Copies addessing info to state data

    sourceBox = source;
    destBox = dest;

    if (sourceBox->type == TERMINAL) {
	// message is from a terminal; reject the message
        printf("ERROR: cheapSend cannot be used for a terminal");
        exit(NOTOK);
    }
    else {
        // message is from a node
        sourceNode = sourceBox->owningNode;

        // Accounts for protocol costs on the sender's CPU
        // times are in milliseconds. KLUDGE: same protocol cost as for
        // messages
        sourceNode->cpu.use(((float) PROTOCOLCOST) / (Mips * 10e3)); 
        if (sourceBox->owningOp != NULL) 
          sourceBox->owningOp->cnt_CPUInstr += PROTOCOLCOST;

        destBox->send((int) this);    // actually send msg to csim process
    }
}

 





///////////////////////////////////////////////////////////////////////////////
//
//  Msg::print     Superclass debug print.  Prints the common message data.
//
///////////////////////////////////////////////////////////////////////////////

void Msg::print(char* traceString) {

  if (Flags[TRACE_MSG]) {
    printf("%s msg %d, type = %d, %d bytes:\n", traceString, 
	   id, type, numBytes);

    if ((int) sourceBox != NULL) 
    {    // if message is addressed
        if (sourceBox->type == TERMINAL)       // from a terminal
	{
	    printf("source is some terminal\n");
/*
	    printf("source terminal = %d\n", 
		((TermMailBox *)sourceBox)->owningTerm->id);
*/
	}
        else  // from a node
	    printf("source node = %d, ", sourceBox->owningNode->id);

        if (destBox->type == TERMINAL)         // to a terminal
	{
	    printf("dest is some terminal\n");
/*
	    printf("dest terminal = %d\n", 
		((TermMailBox*)destBox)->owningTerm->id);
*/
	}
        else  // to a node
		printf("dest node = %d,\n", destBox->owningNode->id);
    }
    else printf("    -- unaddressed --\n");
  };
};

///////////////////////////////////////////////////////////////////////////////
//
//  QueryMsg::QueryMsg Sets message type, and inits other fields to detectably
//                     invalid values.                       
//
///////////////////////////////////////////////////////////////////////////////

QueryMsg::QueryMsg() {
  query = NULL;
  file = NULL;
  index = NULL;
  splitTable = NULL;
  type = QUERYPLAN;
  needMailBox = TRUE;
  nextMsg = NULL;
}

//////////////////////////////////////////////////////////////////////////////
//
//  QueryMsg::alloc     Allocate a QueryMsg.  Just does a new for now.
//
///////////////////////////////////////////////////////////////////////////////

QueryMsg* QueryMsg::alloc() {
  QueryMsg* mp;

  if (freeMsg == NULL) 
  {
	// allocate a new msg
  	mp = new QueryMsg();
  }
  else
  {
	// take one from the front of the free list
	mp = freeMsg;
	freeMsg = freeMsg->nextMsg;
  }
  mp->type = QUERYPLAN;
  mp->query = NULL;

  return(mp);
}

//////////////////////////////////////////////////////////////////////////////
//
//  QueryMsg::free     release a QueryMsg.  Just does a delete for now.
//
///////////////////////////////////////////////////////////////////////////////

void QueryMsg::free(QueryMsg* mp) {

  // put the message back on the free list
  mp->nextMsg = freeMsg;
  freeMsg = mp;
}


///////////////////////////////////////////////////////////////////////////////
//
//  QueryMsg::send      Calculates message size and calls superclass send.
//
///////////////////////////////////////////////////////////////////////////////

void QueryMsg::send(Mailbox* dest, Mailbox* source) {
  numBytes = SIZE_QUERY;
  Msg::send(dest, source);
}

///////////////////////////////////////////////////////////////////////////////
//
//  QueryMsg::print     Debug print.  Prints QueryMsg specific data, then 
//                      calls superclass print.
//
///////////////////////////////////////////////////////////////////////////////

void QueryMsg::print(char* traceString) {
  if (Flags[TRACE_MSG]) {
    Msg::print(traceString);
    //query.print(); ??
  };
}

///////////////////////////////////////////////////////////////////////////////
//
//  CntlMsg::CntlMsg   Sets message type, and inits other fields to detectably
//                     invalid values.                       
//
///////////////////////////////////////////////////////////////////////////////

CntlMsg:: CntlMsg() {
  type = CONTROL;
  opCode = queryOp = -1;
  intData = 0;
  floatData = 0.0;
  splitTable = NULL;
  nextMsg = NULL;
}

//////////////////////////////////////////////////////////////////////////////
//
//  CntlMsg::alloc      Allocate a CntlMsg.  Just does a new for now.
//
///////////////////////////////////////////////////////////////////////////////

CntlMsg* CntlMsg::alloc() {
  CntlMsg* mp;

  if (freeMsg == NULL) 
  {
	// allocate a new msg
  	mp = new CntlMsg();
  }
  else
  {
	// take one from the front of the free list
	mp = freeMsg;
	freeMsg = freeMsg->nextMsg;
  }
  mp->type = CONTROL;
  mp->opCode = -1;
  mp->intData = 0;
  mp->floatData = 0.0;
  mp->splitTable = NULL;
  mp->queryOp = -1;
  mp->numBuckets = 0;
  mp->b1Entries = 0;
  return(mp);
}

//////////////////////////////////////////////////////////////////////////////
//
//  CntlMsg::free     release a CntlMsg.  Just does a delete for now.
//
///////////////////////////////////////////////////////////////////////////////

void CntlMsg::free(CntlMsg* mp) {

// couldn't make the following work. is it necessary??? DEWITT - KLUDGE
// if (mp->splitTable != NULL) delete *(mp->splitTable);

  // put the message back on the free list
  mp->nextMsg = freeMsg;
  freeMsg = mp;
}

///////////////////////////////////////////////////////////////////////////////
//
//  CntlMsg::send       Calculates message size and calls superclass send.
//
///////////////////////////////////////////////////////////////////////////////

void CntlMsg::send(Mailbox* dest, Mailbox* source) {
  numBytes = SIZE_CONTROL;
  Msg::send(dest, source);
}

///////////////////////////////////////////////////////////////////////////////
//
//  CntlMsg::print      Debug print.  Prints CntlMsg specific data, then 
//                      calls superclass print.
//
///////////////////////////////////////////////////////////////////////////////

void CntlMsg::print(char* traceString) {
  if (Flags[TRACE_MSG]) {
    Msg::print(traceString);
    printf("    cntl type = %d,  int data = %d, float data = %f, q op = %d\n",
	   opCode, intData, floatData, queryOp);
    //splitTable.print();
  };
}

///////////////////////////////////////////////////////////////////////////////
//
//  DataMsg::DataMsg   Sets message type, and inits other fields to detectably
//                     invalid values.                       
//
///////////////////////////////////////////////////////////////////////////////

DataMsg:: DataMsg() {
  numTuples = tupleSize = -1;
  type = DATAPKT;
  eofFlag = FALSE;
  nextMsg = NULL;
}

//////////////////////////////////////////////////////////////////////////////
//
//  DataMsg::alloc       Allocate a DataMsg.  Just does a new for now.
//
///////////////////////////////////////////////////////////////////////////////

DataMsg* DataMsg::alloc() {
  DataMsg* mp;

  if (freeMsg == NULL) 
  {
	// allocate a new msg
  	mp = new DataMsg();
  }
  else
  {
	// take one from the front of the free list
	mp = freeMsg;
	freeMsg = freeMsg->nextMsg;
  }
  mp->type = DATAPKT;
  mp->numTuples = -1;
  mp->tupleSize = -1;
  mp->eofFlag = FALSE;
  return(mp);
}

//////////////////////////////////////////////////////////////////////////////
//
//  DataMsg::free     release a DataMsg.  Just does a delete for now.
//
///////////////////////////////////////////////////////////////////////////////

void DataMsg::free(DataMsg* mp) {
  // put the message back on the free list
  mp->nextMsg = freeMsg;
  freeMsg = mp;
}


///////////////////////////////////////////////////////////////////////////////
//
//  DataMsg::print       Debug print.  Prints DataMsg specific data, then 
//                       calls superclass print.
//
///////////////////////////////////////////////////////////////////////////////

void DataMsg::print(char* traceString) {
  if (Flags[TRACE_MSG]) {
    Msg::print(traceString);
    printf("    num tuples = %d, tuple size = %d, eofFlag = %d\n", 
	   numTuples, tupleSize, eofFlag);
  };
}

///////////////////////////////////////////////////////////////////////////////
//
//  DataMsg::send        Calculates message size and calls superclass send.
//
///////////////////////////////////////////////////////////////////////////////

void DataMsg::send(Mailbox* dest, Mailbox* source) {
  numBytes = SIZE_DATAPKT + (numTuples * (tupleSize + TUPLEOVERHEAD));
  if ((numBytes > 10000) || (numBytes < 0))
  printf("datamsg.send being asked to set a message of size %d bytes\n",
	numBytes);
  Msg::send(dest, source);
}






