#include "dbsim.h"
#include "node.h"
#include "split.h"
#include "db.h"
#include "query.h"

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        ldsched.c
//
//  Function:    a left-deep query scheduler
//                 
//  Change Activity:
//
//
///////////////////////////////////////////////////////////////////////////////

extern Node* Nodes[];
extern Mailbox* GlobalSchedPort;
extern int NumDiskNodes;

void startOperator(QueryPlan* , Mailbox*, int);
int getPortIds(QueryPlan*, Mailbox*);
void fillSplitAddrs (QueryPlan*, SplitTable*, int);
SplitTable* copySplitTable(SplitTable*);
void waitDone(QueryPlan*, Mailbox*);
int activate(QueryPlan* , QueryPlan* , Mailbox*, int);
int deactivate(QueryPlan* , Mailbox* );
void sendCntlMsg(QueryPlan* , int, Mailbox* );
void sendSplitTblMsg(QueryPlan* , SplitTable* , Mailbox* );
int ldExec(QueryPlan* , QueryPlan* , QueryMsg* , Mailbox* , int);
int getTblStatus (QueryPlan* , Mailbox* );


// This query scheduler will execute a query plan represented as 
//  a left-deep query .
// node is the node the global query scheduler is running on
// for now we assume that query scheduler is running on the same node

void ldQuerySched(Node* node, QueryMsg* reqMsg, int transId) 
{
    Mailbox* 	qschedMbox;

    create("ldQuerySched");

    if (Flags[TRACE_SCHED]) {
    	printf("query scheduler process started\n");
    	reqMsg->print("query sched got");
    };

    // get a mail mailbox for operators to communicate with
    qschedMbox = new Mailbox("qschedMbox", node);

    // execute the query by calling exec 
    ldExec (reqMsg->query, NULL, reqMsg, qschedMbox, transId);

    // Send "the results" back to the terminal, reusing the request pkt
    reqMsg->send(reqMsg->sourceBox, GlobalSchedPort);

    // should actually clean things up as the operators complete
    delete qschedMbox;
}

// This routine recursively executes the left-deep query tree 
// that is rooted at the current node.  nodes of the tree are activated 
//  via an in-order traversal.

ldExec(QueryPlan* currentNode, QueryPlan* parentNode, QueryMsg* reqMsg, 
	Mailbox* qschedMbox, int transId)
{
    int e;
    JoinPlan* jPlan;
    StorePlan* stPlan;

    printf("ldexec called. opcode = %d\n",currentNode->opCode);

    if (currentNode == NULL) return(OK);

    if (currentNode->opCode == JOIN)
    {
	jPlan = (JoinPlan*) currentNode;

	// execute the left subtree first
	e = ldExec (jPlan->leftChild, currentNode, reqMsg, qschedMbox, transId);
	CKERR_RET("exec left", e);

	// activate/synchronize current and parent nodes
	e = activate(currentNode, parentNode, qschedMbox, transId);
	CKERR_RET("activate", e);

	// execute the right subtree next
	e = ldExec (jPlan->rightChild, currentNode, reqMsg, qschedMbox, transId);
	CKERR_RET("exec right", e);

	e = deactivate(currentNode, qschedMbox);
	CKERR_RET("deactivate", e);

	return(OK);
    }
    else
    if (currentNode->opCode == SELECT)
    {
	// activate/synchronize current and parent nodes
	e = activate(currentNode, parentNode, qschedMbox, transId);
	CKERR_RET("activate", e);

	e = deactivate(currentNode, qschedMbox);
	CKERR_RET("deactivate", e);

	return(OK);
    }

    else
    if (currentNode->opCode == STORE)
    {
	stPlan = (StorePlan*) currentNode;

	// execute the child first
	e = ldExec (stPlan->child, currentNode, reqMsg, qschedMbox, transId);
	CKERR_RET("exec left", e);

	// activate/synchronize current and parent nodes
	e = activate(currentNode, parentNode, qschedMbox, transId);
	CKERR_RET("activate", e);

	e = deactivate(currentNode, qschedMbox);
	CKERR_RET("deactivate", e);

	return(OK);
    }
    else
    {
	printf("unknown upcode = %d in exec\n", currentNode->opCode);
	return (NOTOK);
    }
}

// activate a node in a query tree

activate(QueryPlan* currentNode, QueryPlan* parentNode, Mailbox* qschedMbox,
	int transId)
{
    JoinPlan* 	jPlan;
    QueryPlan* 	lchild;
    QueryPlan*  rchild;
    SplitTable* splitTable;
    int		cnt;

    printf("activate called.  opcode = %d\n",currentNode->opCode);
    // Phase 1, see if the parent operator should be activated frist

    if (parentNode == NULL)  // case 1: current has no parent
    {
	printf("activate:  parentNode is NULL.  \n");
    }
    else
    if ((parentNode->opCode == JOIN) || (parentNode->opCode == STORE))
    {
	jPlan = (JoinPlan*) parentNode;
	if ((currentNode == jPlan->leftChild) || (parentNode->opCode == STORE))
	{
            // Case 2: currentNode is either the left child of a join or 
	    // the child of a store operator.  In this case,
            // start the parent operator (i.e. the join or the store) 

            startOperator(parentNode, qschedMbox, transId); 

            // get the private ports of the various processes involved
            cnt = getPortIds (parentNode, qschedMbox);

	    // build a split table to feed data to the parent operator
	    splitTable = new SplitTable (cnt);
		
    	    // fill in the addresses
    	    fillSplitAddrs (parentNode, splitTable, cnt);

	    if (currentNode->opCode == SELECT)
		((SelectPlan*) currentNode)->splitTable = splitTable;
	    else 
	    if (currentNode->opCode == JOIN)
		((JoinPlan*) currentNode)->splitTable = splitTable;
	    else 
	        printf("in activate, funny opcode=%d\n",currentNode->opCode);
	}
    }

    // Phase 2

    if (currentNode->opCode == SELECT)    // case 1: current is a leaf node
    {
        // start the select operator on all nodes 
        startOperator(currentNode, qschedMbox, transId); 

        // get the private ports of the various operators involved
        cnt = getPortIds (currentNode, qschedMbox);
    }
    else
    if (parentNode == NULL)   // case 2:  parent is NULL
    {
	return(OK);
    }
    else
    if (currentNode->opCode == JOIN)       // case 3: current is a join
    {
	jPlan = (JoinPlan*) currentNode;
	lchild = jPlan->leftChild;
	rchild = jPlan->rightChild;

	if (lchild->opCode == SELECT)
	    splitTable = ((SelectPlan*) lchild)->splitTable;
	else 
	if (lchild->opCode == JOIN) 
	    splitTable = ((JoinPlan*) lchild)->splitTable;

	// make a copy of the left splitTable for right child which
	// must be a select in a left-deep query tree
	((SelectPlan *) rchild)->splitTable = copySplitTable(splitTable);

	// send notify scheduler messages to join processes
	sendCntlMsg(currentNode, HTREQ, qschedMbox);

	// collect replies - end of build phhase
        cnt = getTblStatus (currentNode, qschedMbox);

	// ignore overflow of joins but here is where it is handled
	// in the real gamma system

	// send split tables to current join node
	sendSplitTblMsg(currentNode , ((JoinPlan *)currentNode)->splitTable, 
		qschedMbox);
    }
    else
    {
	// default  - not clear when used
	printf("using default case in activate() - figure out why!!!\n");
	// make split table from active sites of current
	// send split able to current
    }
    return(OK);
}


void startOperator(QueryPlan* qry, Mailbox* qschedMbox, int transId)
{
    Mailbox *nodeMgr;
    QueryMsg *qMsg;
    int	i, cnt;

    // initialize some fields in the actual plan
    qry->activeCnt = 0;
    qry->tupleCnt = 0;
    qry->transId = transId;

    cnt=0;
    for (i=0; i< NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	    // operator is to run on this node 
            // get the mail box of the node manager on the appropriate node 
            nodeMgr = Nodes[i]->nodeMgrPort;
            qMsg = QueryMsg::alloc(); // get query message
            qMsg->query = qry;  
	    qMsg->send(nodeMgr, qschedMbox);
	    cnt++;
	}
    }
    qry->activeCnt = cnt;
}


// get the private ports of all the processes executing an operator
// return count of number of active processes
int getPortIds (QueryPlan* qry, Mailbox* qschedMbox)
{
    int		replyCnt;  // number of responses received
    CntlMsg	*cMsg;
    int		cnt;
    int		nodeId;

    cnt = qry->activeCnt;  // since qry->activeCnt may get modified below
    replyCnt = 0;

    // loop until the message ports of all the operator processes are received
    while (replyCnt < cnt)
    {
	// block waiting for next message
	cMsg = (CntlMsg*) qschedMbox->recv();
	if (cMsg->type != CONTROL)
	    printf("expected CONTROL message in getPortIds. got type=%d\n",
			cMsg->type);
	else
	{
	    switch (cMsg->opCode)
	    {
	    case NEWTASK:
		// update list of active sites
		replyCnt++;
		nodeId = cMsg->intData;  // kind of a kludge but ok
		qry->sites[nodeId] = cMsg->sourceBox; // extract operator's port
	        break;
	    case OPDONE:
	        // This can occur in the case of very short operators
	        qry->activeCnt--;
		qry->tupleCnt += cMsg->intData;
		if (cMsg->queryOp==STORE) 
		{
		    if (cMsg->floatData < ((StorePlan *)qry)->writeTime)
			((StorePlan* )qry)->writeTime = cMsg->floatData;
		}
	        break;
	    default:
	        printf("Unexpected control message type received = %d\n", 
				cMsg->opCode);
		break;
	    }
	    CntlMsg:: free(cMsg);  // release message buffer
	}
    }
    return(replyCnt);
}
	
       

// get end of build phase messages from join processes
// return count of number of active processes
int getTblStatus (QueryPlan* qry, Mailbox* qschedMbox)
{
    int		replyCnt;  // number of responses received
    CntlMsg	*cMsg;

    replyCnt = 0;

    // loop until the message ports of all the operator processes are received
    while (replyCnt < qry->activeCnt)
    {
	// block waiting for next message
	cMsg = (CntlMsg*) qschedMbox->recv();
	if (cMsg->type != CONTROL)
	    printf("expected CONTROL message in getPortIds. got type=%d\n",
			cMsg->type);
	else
	{
	    if (cMsg->opCode == TBLSTATUS) replyCnt++;
	    else printf("Unexpected control message type received = %d\n", 
				cMsg->opCode);
	    CntlMsg:: free(cMsg);  // release message buffer
	}
    }
    return(replyCnt);
}

// Use the mailbox's of the operators executing the parent operator
// to construct a split table for the child operator
// Return the number of entries in the splitTable

void fillSplitAddrs (QueryPlan* qry, SplitTable* splitTable, int size)
{
    // qry points at the parent's query plan
    // splitTable points at the child's split table
    // size is the size of the splitTable

    int		i;
    int		splitCnt;

    splitCnt = 0;
    for (i=0; i  < NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
	    // parent operator is being run on node i
	    splitTable->splitArray[splitCnt].numTuples = 0;
	    splitTable->splitArray[splitCnt].destNode = qry->sites[i];
	    splitCnt++;
	}
    }
    if (splitCnt != size) 
	printf("\n\nmajor problem in fillSplitAddres, size=%d, splitCnt=%d\n\n",		size, splitCnt);
    splitTable->splitCnt = splitCnt;
}

// make a copy of a split table

SplitTable* copySplitTable(SplitTable* from)
{
    int 	i;
    SplitTable*	to;

    to = new SplitTable (from->splitCnt);
    to->splitCnt = from->splitCnt;
    for (i=0;i<from->splitCnt;i++) to->splitArray[i] = from->splitArray[i];
    return(to);
}


// wait for an operator to finish executing.

int deactivate (QueryPlan* qry, Mailbox* qschedMbox)
{
    int		replyCnt;  // number of responses received
    CntlMsg	*cMsg;
    int		cnt;

    cnt = qry->activeCnt;  // since qry->activeCnt may get modified below
    replyCnt = 0;
    printf("deactivating operator %d\n",qry->opCode);

    // join and store queries wait for a notify scheduler message
    // before sending an operator done message

    if ((qry->opCode == JOIN) || (qry->opCode == STORE))
	sendCntlMsg(qry, NOTIFYSCHED, qschedMbox);

    // loop until done  messages have been received from all the operator 
    //  processes 
    while (replyCnt < cnt)
    {
	// block waiting for next message
	cMsg = (CntlMsg*) qschedMbox->recv();

	if ((cMsg->type != CONTROL) || (cMsg->opCode != OPDONE))
	{
	    printf("expected CONTROL message in deactivate. got type=%d\n",
			cMsg->type);
	    return(NOTOK);
	}
	else
	{
	    // got an OPDONE message.
	    replyCnt++;
	    qry->activeCnt--;
	    qry->tupleCnt += cMsg->intData;

	    if (cMsg->queryOp == STORE) 
	    {
		if (cMsg->floatData < ((StorePlan *)qry)->writeTime)
			((StorePlan* )qry)->writeTime = cMsg->floatData;
	    }
	    CntlMsg:: free(cMsg);  // release message buffer
	}
    }
    return(OK);
}


// Send a control message to all processes executing an operator

void sendCntlMsg(QueryPlan* qry, int msgType, Mailbox* qschedMbox)
{
    CntlMsg 	*cMsg;
    int		i;
	
    for (i=0; i < NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
    		cMsg = CntlMsg::alloc(); // get control message
		cMsg->opCode = msgType;  
		cMsg->send (qry->sites[i], qschedMbox);
	}
    }
}

void sendSplitTblMsg(QueryPlan* qry, SplitTable* splitTable,
	Mailbox* qschedMbox)
{
    CntlMsg 	*cMsg;
    int		i;
	
    for (i=0; i < NumDiskNodes; i++) 
    {
	if (qry->activeSites[i] == TRUE)
	{
    		cMsg = CntlMsg::alloc(); // get control message
		cMsg->opCode = NEWSPLIT;  
		cMsg->numBuckets = 1;  // KLUDGE for now hashtab fits in memory
		cMsg->b1Entries = 0;
		// make a copy of the split table
		cMsg->splitTable = copySplitTable(splitTable);
		cMsg->send (qry->sites[i], qschedMbox);
	}
    }
}


