#include <dbsim.h>
#include <cpp_class.h>
#include <object.h>
#include <node.h>
#include <msg.h>
#include <mailbox.h>
#include <terminal.h>
#include <split.h>
#include <op.h>
#include <query.h>
#include <language.h>
#include <db.h>
#include <disk.h>
#include <sched.h>
#include <bufmgr.h>
#include <cost.h>

///////////////////////////////////////////////////////////////////////////////
//
//  Name:        sim.c
//
//  Title:       UW Database Simulator controller
//
//  Function:    Instantiates major sim objects, starts primary sim
//               processes, writes out simulation results, and terminates.
//                 
//  Change Activity:
//
//     08/25/91 (KPB) - Initial version.
//
///////////////////////////////////////////////////////////////////////////////

// Hardware/sofware configuration global variables

int        NumTerminals = 0, // calculated by parser from job class populations
           ActiveTerms = 0,
           Inst_StartIO = INST_STARTIO, // instructions to start an I/O
           NumNodes = DEFAULT_NUMNODES,
           NumDiskNodes = DEFAULT_NUMNODES-1,
           DisksPerNode = DEFAULT_DISKSPERNODE,
           CPUsPerNode =  DEFAULT_CPUSPERNODE,
           KBytesPerNode = DEFAULT_KBYTESPERNODE, // 1K = 1024 bytes (duh!)
           BufSize, // calculated at run time from other config vars
           PagesPerCyl = DEFAULT_PAGESPERCYL,
           PageSize = DEFAULT_PAGESIZE,
           DiskCacheContexts = DEFAULT_DISKCACHECONTEXTS, // disabled if = zero
           DiskCachePages = DEFAULT_DISKCACHEPAGES; // disabled if = zero
float      SeekFactor = DEFAULT_SEEKFACTOR,
           MinLatency = DEFAULT_MINLATENCY,
           DiskRotation = DEFAULT_DISKROTATION,
           SettleTime = DEFAULT_SETTLETIME,
           XferRate = DEFAULT_XFERRATE,
           XferTime, // calculated at run time from other config vars
           Mips =  DEFAULT_MIPS;

// Simulated objects that need to be globally accessible

Mailbox*   GlobalSchedPort;
Node       *Nodes[MAX_NODES] = {NULL};
RandomFile *ConstFiles[MAX_FILES] = {NULL};
File       *Relations[MAX_FILES] = {NULL};
Index      *Indices[MAX_FILES] = {NULL};
Terminal   *Terms[MAX_TERMINALS] = {NULL}; 
JobClass   *JobClasses[MAX_CLASSES] = {NULL};
facility   Network("Network");
MemoryQueue *GlobalQueryWaitQueue ;
int        BufPages[MAX_NODES] = {0};

// Global simulator control data

bool       Flags[MAXFLAGS] = {FALSE},
           CSIMReports = FALSE,
           PlanStats = FALSE,
           Quiet = FALSE,
           PlotMode = FALSE,      // make output easier to grep than to read
           ResetBufferFlag = 0;   // to clear buffer pool after each query
event      SimDone("SimDone");
int	   SimRunning,
           MsgLevel = 0,
           NumClasses = 0,
           SkewFactor = 1,        // KLUDGE for Jeff
           TermIDCounter = 0;
double     TimeWeightedMPL = 0.0; // maintained by scheduler
int        MaxMPL = 0;            // maintained by scheduler

// Global vars used to communicate with SetFlags:

float      PingInterval = 0.0, SimTime = 0.0;  // default is run forever
char       *ConfigFileName = NULL, 
           *SchemaFileName = NULL, 
           *WkldFileName = NULL,
           *SourceFileName = NULL,
           *SimTimeChars;

// Global vars used to communicate with parser/lex:

extern FILE* yyin;        // parser/lex input file
extern int yynerrs;       // parser/lex error count
extern int yylineno;      // lex current line number
SymTable   RealSyms;      // real num symbol table
SymTable   FileSyms;      // file symbol table
SymTable   PlanSyms;      // plan symbol table
SymTable   JobClassSyms;  // job class symbol table

// Function prototypes

void      SetFlags(int argc, char** argv);
void      SetCSIMLimits();
bool      OpenFiles(FILE* &configFile, FILE* &schemaFile, FILE* &wkldFile);
void      OutputConfig();
void      OutputSchema();
void      OutputWorkload();
void      OutputResults();
void      globalScheduler(Node *n);
int       yyparse();
void      WritePing();

///////////////////////////////////////////////////////////////////////////////

extern "C" void sim(int argc, char** argv) {
  int i, waitStat;
  FILE *configFile = NULL, *schemaFile = NULL, *wkldFile = NULL;
  float holdTime;

  create("sim");           // defines itself as "distinguished" CSIM process

  // Sets global flags from the command line.

  SetFlags(argc, argv);
    
  // Verifies language input files were specified and open up OK
    
  if (!OpenFiles(configFile, schemaFile, wkldFile)) return;

  // Parses configuration definition file, possibly resetting config variables

  yyin = configFile;
  SourceFileName = ConfigFileName;
  yylineno = 1;
  yyparse();
  if (Flags[TRACE_PARSER]) printf("config:  yynerrs = %d\n", yynerrs);
  if (yynerrs) return;

  // Calculates any derived configuration constants

  XferTime = ((float) PageSize)/(XferRate * 1000.0);
  BufSize = (int) ((KBytesPerNode * 1024.0) / PageSize); 
  if (!Quiet) OutputConfig();

  // Given the configuration at this point, resets the various CSIM limits.

  SetCSIMLimits();

  // Performs any final initialization needed before cranking up nodes

  Network.set_servicefunc(inf_srv);  // assumes network is NEVER a bottleneck
  SimDone.clear();                   // enables this process to exit gracefully
  SimRunning = TRUE;                 // enables polling processes

  // Instatiates nodes, each will start its own processes

  for (i = 0; i < NumNodes; i++) Nodes[i] = new Node(i);

  // Parses the database definition file

  yyin = schemaFile;
  SourceFileName = SchemaFileName;
  yylineno = 1;
  yyparse();
  if (Flags[TRACE_PARSER]) printf("schema:  yynerrs = %d\n", yynerrs);
  if (yynerrs) return;
  if (!Quiet) OutputSchema();

  // Parses the workload definition file

  yyin = wkldFile;
  SourceFileName = WkldFileName;
  yylineno = 1;
  yyparse();
  if (Flags[TRACE_PARSER]) printf("wkload:  yynerrs = %d\n", yynerrs);
  if (yynerrs) return;
  if (!Quiet) OutputWorkload();

  // Resets CSIM limits again based on additional info in workload file

  SetCSIMLimits();

  // Starts a global query scheduler process on node "NumNodes"

  if (Flags[TRACE_CONTROL]) 
    printf("starting the global scheduler node on node %d\n", NumNodes - 1);
  if (Flags[TRACE_CONTROL]) 
    printf("queries run on nodes 0 to %d\n",NumNodes-2);

  ActiveTerms = NumTerminals;
  GlobalSchedPort = new Mailbox("GlobalSchedPort", Nodes[NumNodes-1]);
  globalScheduler(Nodes[NumNodes-1]);

  // instantiates a global query queue to be used by the scheduler
  GlobalQueryWaitQueue = new MemoryQueue();

  // initialize the buffer array to be used by the scheduler
  for(i = 0; i < NumNodes; i++) BufPages[i] = BufSize;

  // Instantiates/starts up terminal objects/processes for each job class

  for (i = 0; i < NumClasses; i++) JobClasses[i]->start();

  // Waits for the simulation to finish, writes simulation output and exits.
  // Writes out periodic progress msgs if -p flag was specified.

  if (SimTime != 0.0) {                  // if user specified a simulation time
    fprintf(stderr, "\n%s:  specified simulation time:  %s (%.0f msecs)\n", 
	   SIMNAME, SimTimeChars, SimTime);

    if (PingInterval)		         // user wants periodic progress msgs
      do {
	if (simtime()+PingInterval > SimTime){ // should we quit b4 next ping?
	  holdTime = SimTime - simtime();
	  if (holdTime < 0.0) holdTime = 0.0; // cant be too careful!
        }				      
	else                             // can fit next ping in b4 sim done
	  holdTime = PingInterval;
	hold(holdTime);                  // wait till next interval
	WritePing();
      } while (simtime() < SimTime);     // keep pinging until time up
	
    else                                 // no ping msgs, just wait quietly
      hold(SimTime);

    fprintf(stderr, "\n%s:  simulation time expired\n\n", SIMNAME);
    SimRunning = FALSE;                  // disables polling processes 

  // No simulation time parm.  Stops with cntl C or all terminals having finite
  // plan sequences, or a "primary" job class exists.

  } else {
    fprintf(stderr, "\n%s:  no specified simulation time limit\n", SIMNAME);

    if (PingInterval)
      do {                               // until no more active terminals
        waitStat = SimDone.timed_wait(PingInterval);
	WritePing();
      } while (waitStat != EVENT_OCCURRED); // loops while still active terms 

    else                                 // no ping msgs, just wait quietly
      SimDone.wait();

    SimRunning = FALSE;                  // disables polling processes
  } 

  OutputResults();
}

///////////////////////////////////////////////////////////////////////////////

bool OpenFiles(FILE* &configFile, FILE* &schemaFile, FILE* &wkldFile) {

  if (ConfigFileName == NULL ||
      SchemaFileName == NULL ||
      WkldFileName == NULL) {
    printf("usage:  %s -c config_file -s schema_file -w wkld_file -t timespec -d flags\n                -q(uiet) -p(ing) -C(SIM reports) -P(lot mode)\n\nwhere:  - timespec is one of ss, mm:ss, or hh:mm:ss in simulated time\n        - flags is one of\n%s%s%s%s%s%s%s%s%s%s%s%s",
	   SIMNAME,
           "          B - Buffer Manager\n",
           "          D - Disk\n",
           "          I - I/O\n",
           "          T - Terminals\n",
           "          S - Scheduler\n",
           "          O - Operators\n",
           "          M - Messages\n",
           "          L - Lock Manager\n",
           "          Q - Queries\n",
           "          P - Parser\n",
           "          C - General control flow\n",
           "          A - All traces\n");
    return(FALSE);
  }

  configFile = fopen(ConfigFileName, "r");
  if (configFile == NULL) {
    printf("%s:  error attempting to open configuration definition file %s\n", 
	   SIMNAME, ConfigFileName);
    return(FALSE);
  }

  schemaFile = fopen(SchemaFileName, "r");
  if (schemaFile == NULL) {
    printf("%s:  error attempting to open schema definition file %s\n", 
	   SIMNAME, SchemaFileName);
    return(FALSE);
  }

  wkldFile = fopen(WkldFileName, "r");
  if (wkldFile == NULL) {
    printf("%s:  error attempting to open workload definition file %s\n", 
	   SIMNAME, WkldFileName);
    return(FALSE);
  }

  return(TRUE);
}

///////////////////////////////////////////////////////////////////////////////

void SetCSIMLimits(){
  int i;

  // 1 per max disks per node, one per node for cpu, 1 for network, 9 to spare
  i = max_facilities (NumNodes * (1 + MAX_DISKSPERNODE) + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM facilities set to %d\n",i);

  // 1 per max disks per node, one per cpu per node, 1 for network, 9 to spare
  i = max_servers(NumNodes * (CPUsPerNode + MAX_DISKSPERNODE) + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM servers set to %d\n",i);

  // 1 per max disks per node, one per node for cpu, 1 for network, 9 to spare
  i = max_qtables (NumNodes * (1 + MAX_DISKSPERNODE) + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM qtables set to %d\n",i);

  // 1 per max disks per node + 11 per max job classes + 8*20 for plan stats
  i = max_tables (NumNodes * MAX_DISKSPERNODE + 11 * MAX_CLASSES + 8 * 20);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM tables set to %d\n",i);

  // one histogram for each facility???
  i = max_histograms (NumNodes * (1 + MAX_DISKSPERNODE) + 10);  
  if (Flags[TRACE_CONTROL]) printf("number of CSIM histograms set to %d\n",i);

  // one mailbox for each operator on each node, plus global scheduler,
  // global deadlock detector, 1000 query schedulers (max MPL), and terminals
  i = max_mailboxes (NumNodes * (MAX_OPSPERNODE + 5) + NumTerminals + 1000);
  if (Flags[TRACE_CONTROL]) printf("number of CSIM mailboxes set to %d\n",i);

  // 2 per disk per node (assume 30 I/Os in queue), 1 per buf page per node 
  // plus one per mailbox (processes wait on a recv when mb is empty)
  i = max_events (NumNodes*(2 * DisksPerNode * 30 + BufSize)+max_mailboxes(0));
  if (Flags[TRACE_CONTROL]) printf("number of CSIM events set to %d\n",i);

  // one mailbox for each operator on each node, plus global scheduler,
  // global deadlock detector, 1000 query schedulers (max MPL), and terminals
  i = max_processes (NumNodes * (MAX_OPSPERNODE + 5) + NumTerminals + 1000);
  if (Flags[TRACE_CONTROL]) printf("number of CSIM processes set to %d\n",i);

  // assume each operator on each node is sending messages to
  // operators on each of the other nodes, and a backlog of 1000 msgs max
  i = max_messages (1000*NumNodes * (MAX_OPSPERNODE * NumNodes) + NumTerminals);
  if (Flags[TRACE_CONTROL]) printf("number of CSIM messages set to %d\n",i);

}

///////////////////////////////////////////////////////////////////////////////

void OutputConfig(){
  char* diskCacheStat = "";

  if (DiskCachePages == 0 || DiskCacheContexts == 0) 
    diskCacheStat = "(disabled)";

  printf("\n Hardware Configuration\n -------- -------------\n  %7d nodes (1 host, %d with disks)\n  %7d disks/node\n  %7d processors/node\n  %7.2f MIPS/processor\n  %7d Kbytes RAM/node\n  %7d byte pages\n\n  %7d pages/cylinder\n  %7.3f msecs min rotational latency\n  %7.3f msecs/rotation\n  %7.3f msecs settle time\n  %7.3f seek factor\n  %7.3f Mbytes/sec transfer rate\n  %7.3f msecs/page transfer time\n  %7d disk cache pages per context %s\n  %7d disk cache contexts %s\n\n  %7d instr to start an I/O\n\n",
	 NumNodes, NumNodes-1,
	 DisksPerNode, 
	 CPUsPerNode, 
	 Mips, 
	 KBytesPerNode,
	 PageSize,
	 PagesPerCyl,
	 MinLatency,
	 DiskRotation,
	 SettleTime,
	 SeekFactor,
	 XferRate,
	 XferTime,
	 DiskCachePages,
	 diskCacheStat,
	 DiskCacheContexts,
	 diskCacheStat,
	 Inst_StartIO);
}

///////////////////////////////////////////////////////////////////////////////

void OutputSchema(){}

///////////////////////////////////////////////////////////////////////////////

void OutputWorkload(){}

///////////////////////////////////////////////////////////////////////////////

void OutputResults(){
  int i;

  printf("\n\nElaspsed simulated time:  %f msecs\n", simtime());

  // Writes out global scheduler statistics

  printf("\n\nScheduling Statistics:\n----------------------\n\n");
  printf("\tAverageMPL\t%.2f\n\tMaxMPL\t\t%d\n", 
	 (TimeWeightedMPL / simtime()), MaxMPL);
  printf("\n\nGlobal Memory Wait Queue Statistics:\n------------------------------------\n");
  GlobalQueryWaitQueue->printStats();

  // Writes out job class related statistics for each class

  printf("\n\nJob Class Statistics:\n---------------------\n");
  for (i = 0; i < NumClasses; i++) JobClasses[i]->writeStats();

  // If one or more plans have been marked for statistics gathering, prints
  // out the stats for each one.

  if (PlanStats) {
    printf("\n\nPlan Statistics:\n---------------\n");
    PlanSyms.writeStats();
  }
    
  if (!Quiet) {

    // Writes out CPU related statistics for each node

    printf("\n\nCPU Statistics:\n---------------\n\n");
    printf("  Node   util   thruput  response  q length  svc time     compl\n");
    printf("  ----  -----  --------  --------  --------  --------  --------\n");
    for (i = 0; i < NumNodes; i++) Nodes[i]->printCPUStats();

    // Writes out buffer related statistics for each node

    printf("\n\nBuffer Statistics:\n------------------\n\n");
    printf("  Node  hit %%      hits      refs    writes     reads\n");
    printf("  ----  -----  --------  --------  --------  --------\n");
    for (i = 0; i < NumNodes; i++) Nodes[i]->bufMgr->PrintBufferPoolStats();
    
    // Writes out disk related statistics for each node
      
    printf("\n\nDisk Statistics:\n----------------\n\n");
    printf("              arm    resp avg svc min svc max svc std dev cache   cntlr     arm   avg  max\n");
    printf("  Node disk  util @ cntlr  @ disk  @ disk  @ disk  @ disk hit %%   compl   compl  qlen qlen\n");
    printf("  ---- ---- ----- ------- ------- ------- ------- ------- ----- ------- ------- ----- ----\n");
    for (i = 0; i < NumNodes; i++) Nodes[i]->printDiskStats();

  }
  
  if (CSIMReports) report();             // writes out CSIM standard reports
  if (!Quiet) mdlstat();                 // prints resources used by simulation

}

///////////////////////////////////////////////////////////////////////////////

void      WritePing(){
  int i, totCompletions = 0;
  char primClassMsg[80];

  primClassMsg[0] = '\0';

  for (i = 0; i < NumClasses; i++) {
    totCompletions += JobClasses[i]->completions;
    if (JobClasses[i]->stopCount >= 0) // primary job class?
      sprintf(primClassMsg, "(%8d for primary job class)",
	      JobClasses[i]->completions);
  }
  
  fprintf(stderr, "%s: time %10.2f, %8d completions %s\n",
	  SIMNAME, simtime(), totCompletions, primClassMsg);
}









